@@ -2,8 +2,8 @@
 /*
  * Mandos plugin runner - Run Mandos plugins
  *
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -23,38 +23,37 @@
  */
 
 #define _GNU_SOURCE		/* TEMP_FAILURE_RETRY(), getline(),
-				   asprintf(), O_CLOEXEC */
+				   O_CLOEXEC, pipe2() */
 #include <stddef.h>		/* size_t, NULL */
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
 #include <stdio.h>		/* fileno(), fprintf(),
-				   stderr, STDOUT_FILENO */
-#include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
-				   stat, waitpid(), WIFEXITED(),
-				   WEXITSTATUS(), wait(), pid_t,
-				   uid_t, gid_t, getuid(), getgid(),
-				   dirfd() */
+				   stderr, STDOUT_FILENO, fclose() */
+#include <sys/types.h>	        /* fstat(), struct stat, waitpid(),
+				   WIFEXITED(), WEXITSTATUS(), wait(),
+				   pid_t, uid_t, gid_t, getuid(),
+				   getgid() */
 #include <sys/select.h>		/* fd_set, select(), FD_ZERO(),
 				   FD_SET(), FD_ISSET(), FD_CLR */
 #include <sys/wait.h>		/* wait(), waitpid(), WIFEXITED(),
 				   WEXITSTATUS(), WTERMSIG(),
 				   WCOREDUMP() */
-#include <sys/stat.h>		/* struct stat, stat(), S_ISREG() */
+#include <sys/stat.h>		/* struct stat, fstat(), S_ISREG() */
 #include <iso646.h>		/* and, or, not */
-#include <dirent.h>		/* DIR, struct dirent, fdopendir(),
-				   readdir(), closedir(), dirfd() */
-#include <unistd.h>		/* struct stat, stat(), S_ISREG(),
-				   fcntl(), setuid(), setgid(),
-				   F_GETFD, F_SETFD, FD_CLOEXEC,
-				   access(), pipe(), fork(), close()
-				   dup2(), STDOUT_FILENO, _exit(),
-				   execv(), write(), read(),
-				   close() */
+#include <dirent.h>		/* struct dirent, scandirat() */
+#include <unistd.h>		/* fcntl(), F_GETFD, F_SETFD,
+				   FD_CLOEXEC, write(), STDOUT_FILENO,
+				   struct stat, fstat(), close(),
+				   setgid(), setuid(), S_ISREG(),
+				   faccessat() pipe2(), fork(),
+				   _exit(), dup2(), fexecve(), read()
+				*/
 #include <fcntl.h>		/* fcntl(), F_GETFD, F_SETFD,
-				   FD_CLOEXEC */
-#include <string.h>		/* strsep, strlen(), asprintf(),
-				   strsignal(), strcmp(), strncmp() */
+				   FD_CLOEXEC, openat(), scandirat(),
+				   pipe2() */
+#include <string.h>		/* strsep, strlen(), strsignal(),
+				   strcmp(), strncmp() */
 #include <errno.h>		/* errno */
 #include <argp.h>		/* struct argp_option, struct
 				   argp_state, struct argp,
@@ -72,6 +71,7 @@
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
 #include <errno.h> 		/* errno */
 #include <error.h>		/* error() */
+#include <fnmatch.h>		/* fnmatch() */
 
 #define BUFFER_SIZE 256
 
@@ -105,6 +105,7 @@ static plugin *plugin_list = NULL;
 
 /* Gets an existing plugin based on name,
    or if none is found, creates a new one */
+__attribute__((warn_unused_result))
 static plugin *getplugin(char *name){
   /* Check for existing plugin with that name */
   for(plugin *p = plugin_list; p != NULL; p = p->next){
@@ -171,18 +172,20 @@ static plugin *getplugin(char *name){
 }
 
 /* Helper function for add_argument and add_environment */
-__attribute__((nonnull))
+__attribute__((nonnull, warn_unused_result))
 static bool add_to_char_array(const char *new, char ***array,
 			      int *len){
   /* Resize the pointed-to array to hold one more pointer */
+  char **new_array = NULL;
   do {
-    *array = realloc(*array, sizeof(char *)
-		     * (size_t) ((*len) + 2));
-  } while(*array == NULL and errno == EINTR);
+    new_array = realloc(*array, sizeof(char *)
+			* (size_t) ((*len) + 2));
+  } while(new_array == NULL and errno == EINTR);
   /* Malloc check */
-  if(*array == NULL){
+  if(new_array == NULL){
     return false;
   }
+  *array = new_array;
   /* Make a copy of the new string */
   char *copy;
   do {
@@ -200,7 +203,7 @@ static bool add_to_char_array(const char *new, char ***array,
 }
 
 /* Add to a plugin's argument vector */
-__attribute__((nonnull(2)))
+__attribute__((nonnull(2), warn_unused_result))
 static bool add_argument(plugin *p, const char *arg){
   if(p == NULL){
     return false;
@@ -209,7 +212,7 @@ static bool add_argument(plugin *p, const char *arg){
 }
 
 /* Add to a plugin's environment */
-__attribute__((nonnull(2)))
+__attribute__((nonnull(2), warn_unused_result))
 static bool add_environment(plugin *p, const char *def, bool replace){
   if(p == NULL){
     return false;
@@ -217,19 +220,19 @@ static bool add_environment(plugin *p, const char *def, bool replace){
   /* namelen = length of name of environment variable */
   size_t namelen = (size_t)(strchrnul(def, '=') - def);
   /* Search for this environment variable */
-  for(char **e = p->environ; *e != NULL; e++){
-    if(strncmp(*e, def, namelen + 1) == 0){
+  for(char **envdef = p->environ; *envdef != NULL; envdef++){
+    if(strncmp(*envdef, def, namelen + 1) == 0){
       /* It already exists */
       if(replace){
-	char *new;
+	char *new_envdef;
 	do {
-	  new = realloc(*e, strlen(def) + 1);
-	} while(new == NULL and errno == EINTR);
-	if(new == NULL){
+	  new_envdef = realloc(*envdef, strlen(def) + 1);
+	} while(new_envdef == NULL and errno == EINTR);
+	if(new_envdef == NULL){
 	  return false;
 	}
-	*e = new;
-	strcpy(*e, def);
+	*envdef = new_envdef;
+	strcpy(*envdef, def);
       }
       return true;
     }
@@ -237,11 +240,13 @@ static bool add_environment(plugin *p, const char *def, bool replace){
   return add_to_char_array(def, &(p->environ), &(p->envc));
 }
 
+#ifndef O_CLOEXEC
 /*
  * Based on the example in the GNU LibC manual chapter 13.13 "File
  * Descriptor Flags".
  | [[info:libc:Descriptor%20Flags][File Descriptor Flags]] |
  */
+__attribute__((warn_unused_result))
 static int set_cloexec_flag(int fd){
   int ret = (int)TEMP_FAILURE_RETRY(fcntl(fd, F_GETFD, 0));
   /* If reading the flags failed, return error indication now. */
@@ -252,6 +257,7 @@ static int set_cloexec_flag(int fd){
   return (int)TEMP_FAILURE_RETRY(fcntl(fd, F_SETFD,
 				       ret | FD_CLOEXEC));
 }
+#endif	/* not O_CLOEXEC */
 
 
 /* Mark processes as completed when they exit, and save their exit
@@ -289,7 +295,7 @@ static void handle_sigchld(__attribute__((unused)) int sig){
 }
 
 /* Prints out a password to stdout */
-__attribute__((nonnull))
+__attribute__((nonnull, warn_unused_result))
 static bool print_out_password(const char *buffer, size_t length){
   ssize_t ret;
   for(size_t written = 0; written < length; written += (size_t)ret){
@@ -343,9 +349,7 @@ int main(int argc, char *argv[]){
   char *plugindir = NULL;
   char *argfile = NULL;
   FILE *conffp;
-  size_t d_name_len;
-  DIR *dir = NULL;
-  struct dirent *dirst;
+  struct dirent **direntries = NULL;
   struct stat st;
   fd_set rfds_all;
   int ret, maxfd = 0;
@@ -359,6 +363,7 @@ int main(int argc, char *argv[]){
 				      .sa_flags = SA_NOCLDSTOP };
   char **custom_argv = NULL;
   int custom_argc = 0;
+  int dir_fd = -1;
   
   /* Establish a signal handler */
   sigemptyset(&sigchld_action.sa_mask);
@@ -435,10 +440,13 @@ int main(int argc, char *argv[]){
 	    break;
 	  }
 	}
+	errno = 0;
       }
       break;
     case 'G':			/* --global-env */
-      add_environment(getplugin(NULL), arg, true);
+      if(add_environment(getplugin(NULL), arg, true)){
+	errno = 0;
+      }
       break;
     case 'o':			/* --options-for */
       {
@@ -461,6 +469,7 @@ int main(int argc, char *argv[]){
 	    break;
 	  }
 	}
+	errno = 0;
       }
       break;
     case 'E':			/* --env-for */
@@ -478,7 +487,9 @@ int main(int argc, char *argv[]){
 	  errno = EINVAL;
 	  break;
 	}
-	add_environment(getplugin(arg), envdef, true);
+	if(add_environment(getplugin(arg), envdef, true)){
+	  errno = 0;
+	}
       }
       break;
     case 'd':			/* --disable */
@@ -486,6 +497,7 @@ int main(int argc, char *argv[]){
 	plugin *p = getplugin(arg);
 	if(p != NULL){
 	  p->disabled = true;
+	  errno = 0;
 	}
       }
       break;
@@ -494,12 +506,16 @@ int main(int argc, char *argv[]){
 	plugin *p = getplugin(arg);
 	if(p != NULL){
 	  p->disabled = false;
+	  errno = 0;
 	}
       }
       break;
     case 128:			/* --plugin-dir */
       free(plugindir);
       plugindir = strdup(arg);
+      if(plugindir != NULL){
+	errno = 0;
+      }
       break;
     case 129:			/* --config-file */
       /* This is already done by parse_opt_config_file() */
@@ -513,6 +529,7 @@ int main(int argc, char *argv[]){
 	break;
       }
       uid = (uid_t)tmp_id;
+      errno = 0;
       break;
     case 131:			/* --groupid */
       tmp_id = strtoimax(arg, &tmp, 10);
@@ -523,6 +540,7 @@ int main(int argc, char *argv[]){
 	break;
       }
       gid = (gid_t)tmp_id;
+      errno = 0;
       break;
     case 132:			/* --debug */
       debug = true;
@@ -576,6 +594,9 @@ int main(int argc, char *argv[]){
     case 129:			/* --config-file */
       free(argfile);
       argfile = strdup(arg);
+      if(argfile != NULL){
+	errno = 0;
+      }
       break;
     case 130:			/* --userid */
     case 131:			/* --groupid */
@@ -664,13 +685,19 @@ int main(int argc, char *argv[]){
 	}
 	
 	custom_argc += 1;
-	custom_argv = realloc(custom_argv, sizeof(char *)
-			      * ((unsigned int) custom_argc + 1));
-	if(custom_argv == NULL){
-	  error(0, errno, "realloc");
-	  exitstatus = EX_OSERR;
-	  free(org_line);
-	  goto fallback;
+	{
+	  char **new_argv = realloc(custom_argv, sizeof(char *)
+				    * ((unsigned int)
+				       custom_argc + 1));
+	  if(new_argv == NULL){
+	    error(0, errno, "realloc");
+	    exitstatus = EX_OSERR;
+	    free(new_arg);
+	    free(org_line);
+	    goto fallback;
+	  } else {
+	    custom_argv = new_argv;
+	  }
 	}
 	custom_argv[custom_argc-1] = new_arg;
 	custom_argv[custom_argc] = NULL;
@@ -753,7 +780,9 @@ int main(int argc, char *argv[]){
        <http://bugs.debian.org/633582> */
     int plugindir_fd = open(/* plugindir or */ PDIR, O_RDONLY);
     if(plugindir_fd == -1){
-      error(0, errno, "open");
+      if(errno != ENOENT){
+	error(0, errno, "open(\"" PDIR "\")");
+      }
     } else {
       ret = (int)TEMP_FAILURE_RETRY(fstat(plugindir_fd, &st));
       if(ret == -1){
@@ -771,7 +800,7 @@ int main(int argc, char *argv[]){
   }
   
   /* Lower permissions */
-  setgid(gid);
+  ret = setgid(gid);
   if(ret == -1){
     error(0, errno, "setgid");
   }
@@ -782,24 +811,13 @@ int main(int argc, char *argv[]){
   
   /* Open plugin directory with close_on_exec flag */
   {
-    int dir_fd = -1;
-    if(plugindir == NULL){
-      dir_fd = open(PDIR, O_RDONLY |
+    dir_fd = open(plugindir != NULL ? plugindir : PDIR, O_RDONLY |
 #ifdef O_CLOEXEC
-		    O_CLOEXEC
+		  O_CLOEXEC
 #else  /* not O_CLOEXEC */
-		    0
+		  0
 #endif	/* not O_CLOEXEC */
-		    );
-    } else {
-      dir_fd = open(plugindir, O_RDONLY |
-#ifdef O_CLOEXEC
-		    O_CLOEXEC
-#else  /* not O_CLOEXEC */
-		    0
-#endif	/* not O_CLOEXEC */
-		    );
-    }
+		  );
     if(dir_fd == -1){
       error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
@@ -811,129 +829,104 @@ int main(int argc, char *argv[]){
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
-      TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
     }
 #endif	/* O_CLOEXEC */
-    
-    dir = fdopendir(dir_fd);
-    if(dir == NULL){
-      error(0, errno, "Could not open plugin dir");
-      TEMP_FAILURE_RETRY(close(dir_fd));
-      exitstatus = EX_OSERR;
-      goto fallback;
+  }
+  
+  int good_name(const struct dirent * const dirent){
+    const char * const patterns[] = { ".*", "#*#", "*~", "*.dpkg-new",
+				      "*.dpkg-old", "*.dpkg-bak",
+				      "*.dpkg-divert", NULL };
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wcast-qual"
+#endif
+    for(const char **pat = (const char **)patterns;
+	*pat != NULL; pat++){
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
+      if(fnmatch(*pat, dirent->d_name, FNM_FILE_NAME | FNM_PERIOD)
+	 != FNM_NOMATCH){
+	if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " matching pattern %s\n", dirent->d_name, *pat);
+	}
+	return 0;
+      }
     }
+    return 1;
+  }
+  
+#ifdef __GLIBC__
+#if __GLIBC_PREREQ(2, 15)
+  int numplugins = scandirat(dir_fd, ".", &direntries, good_name,
+			     alphasort);
+#else  /* not __GLIBC_PREREQ(2, 15) */
+  int numplugins = scandir(plugindir != NULL ? plugindir : PDIR,
+			   &direntries, good_name, alphasort);
+#endif	/* not __GLIBC_PREREQ(2, 15) */
+#else	/* not __GLIBC__ */
+  int numplugins = scandir(plugindir != NULL ? plugindir : PDIR,
+			   &direntries, good_name, alphasort);
+#endif	/* not __GLIBC__ */
+  if(numplugins == -1){
+    error(0, errno, "Could not scan plugin dir");
+    direntries = NULL;
+    exitstatus = EX_OSERR;
+    goto fallback;
   }
   
   FD_ZERO(&rfds_all);
   
   /* Read and execute any executable in the plugin directory*/
-  while(true){
-    do {
-      dirst = readdir(dir);
-    } while(dirst == NULL and errno == EINTR);
+  for(int i = 0; i < numplugins; i++){
     
-    /* All directory entries have been processed */
-    if(dirst == NULL){
-      if(errno == EBADF){
-	error(0, errno, "readdir");
-	exitstatus = EX_IOERR;
-	goto fallback;
-      }
-      break;
-    }
-    
-    d_name_len = strlen(dirst->d_name);
-    
-    /* Ignore dotfiles, backup files and other junk */
-    {
-      bool bad_name = false;
-      
-      const char const *bad_prefixes[] = { ".", "#", NULL };
-      
-      const char const *bad_suffixes[] = { "~", "#", ".dpkg-new",
-					   ".dpkg-old",
-					   ".dpkg-bak",
-					   ".dpkg-divert", NULL };
-      for(const char **pre = bad_prefixes; *pre != NULL; pre++){
-	size_t pre_len = strlen(*pre);
-	if((d_name_len >= pre_len)
-	   and strncmp((dirst->d_name), *pre, pre_len) == 0){
-	  if(debug){
-	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		    " with bad prefix %s\n", dirst->d_name, *pre);
-	  }
-	  bad_name = true;
-	  break;
-	}
-      }
-      if(bad_name){
-	continue;
-      }
-      for(const char **suf = bad_suffixes; *suf != NULL; suf++){
-	size_t suf_len = strlen(*suf);
-	if((d_name_len >= suf_len)
-	   and (strcmp((dirst->d_name) + d_name_len-suf_len, *suf)
-		== 0)){
-	  if(debug){
-	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		    " with bad suffix %s\n", dirst->d_name, *suf);
-	  }
-	  bad_name = true;
-	  break;
-	}
-      }
-      
-      if(bad_name){
-	continue;
-      }
-    }
-    
-    char *filename;
-    if(plugindir == NULL){
-      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, PDIR "/%s",
-					     dirst->d_name));
-    } else {
-      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, "%s/%s",
-					     plugindir,
-					     dirst->d_name));
-    }
-    if(ret < 0){
-      error(0, errno, "asprintf");
+    int plugin_fd = openat(dir_fd, direntries[i]->d_name, O_RDONLY);
+    if(plugin_fd == -1){
+      error(0, errno, "Could not open plugin");
+      free(direntries[i]);
       continue;
     }
-    
-    ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
+    ret = (int)TEMP_FAILURE_RETRY(fstat(plugin_fd, &st));
     if(ret == -1){
       error(0, errno, "stat");
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
+      free(direntries[i]);
       continue;
     }
     
     /* Ignore non-executable files */
     if(not S_ISREG(st.st_mode)
-       or (TEMP_FAILURE_RETRY(access(filename, X_OK)) != 0)){
+       or (TEMP_FAILURE_RETRY(faccessat(dir_fd, direntries[i]->d_name,
+					X_OK, 0)) != 0)){
       if(debug){
-	fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		" with bad type or mode\n", filename);
+	fprintf(stderr, "Ignoring plugin dir entry \"%s/%s\""
+		" with bad type or mode\n",
+		plugindir != NULL ? plugindir : PDIR,
+		direntries[i]->d_name);
       }
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
+      free(direntries[i]);
       continue;
     }
     
-    plugin *p = getplugin(dirst->d_name);
+    plugin *p = getplugin(direntries[i]->d_name);
     if(p == NULL){
       error(0, errno, "getplugin");
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
+      free(direntries[i]);
       continue;
     }
     if(p->disabled){
       if(debug){
 	fprintf(stderr, "Ignoring disabled plugin \"%s\"\n",
-		dirst->d_name);
+		direntries[i]->d_name);
       }
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
+      free(direntries[i]);
       continue;
     }
     {
@@ -953,9 +946,8 @@ int main(int argc, char *argv[]){
 	}
       }
     }
-    /* If this plugin has any environment variables, we will call
-       using execve and need to duplicate the environment from this
-       process, too. */
+    /* If this plugin has any environment variables, we need to
+       duplicate the environment from this process, too. */
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
@@ -965,25 +957,47 @@ int main(int argc, char *argv[]){
     }
     
     int pipefd[2];
+#ifndef O_CLOEXEC
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
+#else  /* O_CLOEXEC */
+    ret = (int)TEMP_FAILURE_RETRY(pipe2(pipefd, O_CLOEXEC));
+#endif	/* O_CLOEXEC */
     if(ret == -1){
       error(0, errno, "pipe");
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
+    if(pipefd[0] >= FD_SETSIZE){
+      fprintf(stderr, "pipe()[0] (%d) >= FD_SETSIZE (%d)", pipefd[0],
+	      FD_SETSIZE);
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
+      exitstatus = EX_OSERR;
+      free(direntries[i]);
+      goto fallback;
+    }
+#ifndef O_CLOEXEC
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
+#endif	/* not O_CLOEXEC */
     /* Block SIGCHLD until process is safely in process list */
     ret = (int)TEMP_FAILURE_RETRY(sigprocmask(SIG_BLOCK,
 					      &sigchld_action.sa_mask,
@@ -991,6 +1005,7 @@ int main(int argc, char *argv[]){
     if(ret < 0){
       error(0, errno, "sigprocmask");
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
     /* Starting a new process to be watched */
@@ -1000,7 +1015,12 @@ int main(int argc, char *argv[]){
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
       error(0, errno, "fork");
+      TEMP_FAILURE_RETRY(sigprocmask(SIG_UNBLOCK,
+				     &sigchld_action.sa_mask, NULL));
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
     if(pid == 0){
@@ -1022,29 +1042,20 @@ int main(int argc, char *argv[]){
 	_exit(EX_OSERR);
       }
       
-      if(dirfd(dir) < 0){
-	/* If dir has no file descriptor, we could not set FD_CLOEXEC
-	   above and must now close it manually here. */
-	closedir(dir);
-      }
-      if(p->environ[0] == NULL){
-	if(execv(filename, p->argv) < 0){
-	  error(0, errno, "execv for %s", filename);
-	  _exit(EX_OSERR);
-	}
-      } else {
-	if(execve(filename, p->argv, p->environ) < 0){
-	  error(0, errno, "execve for %s", filename);
-	  _exit(EX_OSERR);
-	}
+      if(fexecve(plugin_fd, p->argv,
+		(p->environ[0] != NULL) ? p->environ : environ) < 0){
+	error(0, errno, "fexecve for %s/%s",
+	      plugindir != NULL ? plugindir : PDIR,
+	      direntries[i]->d_name);
+	_exit(EX_OSERR);
       }
       /* no return */
     }
     /* Parent process */
     TEMP_FAILURE_RETRY(close(pipefd[1])); /* Close unused write end of
 					     pipe */
-    free(filename);
-    plugin *new_plugin = getplugin(dirst->d_name);
+    TEMP_FAILURE_RETRY(close(plugin_fd));
+    plugin *new_plugin = getplugin(direntries[i]->d_name);
     if(new_plugin == NULL){
       error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
@@ -1054,8 +1065,10 @@ int main(int argc, char *argv[]){
         error(0, errno, "sigprocmask");
       }
       exitstatus = EX_OSERR;
+      free(direntries[i]);
       goto fallback;
     }
+    free(direntries[i]);
     
     new_plugin->pid = pid;
     new_plugin->fd = pipefd[0];
@@ -1071,16 +1084,30 @@ int main(int argc, char *argv[]){
       goto fallback;
     }
     
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
     FD_SET(new_plugin->fd, &rfds_all); /* Spurious warning from
-					  -Wconversion */
+					  -Wconversion in GNU libc
+					  before 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
     
     if(maxfd < new_plugin->fd){
       maxfd = new_plugin->fd;
     }
   }
   
-  TEMP_FAILURE_RETRY(closedir(dir));
-  dir = NULL;
+  free(direntries);
+  direntries = NULL;
+  TEMP_FAILURE_RETRY(close(dir_fd));
+  dir_fd = -1;
   free_plugin(getplugin(NULL));
   
   for(plugin *p = plugin_list; p != NULL; p = p->next){
@@ -1132,8 +1159,20 @@ int main(int argc, char *argv[]){
 	  }
 	  
 	  /* Remove the plugin */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
 	  FD_CLR(proc->fd, &rfds_all); /* Spurious warning from
-					  -Wconversion */
+					  -Wconversion in GNU libc
+					  before 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
 	  
 	  /* Block signal while modifying process_list */
 	  ret = (int)TEMP_FAILURE_RETRY(sigprocmask
@@ -1179,22 +1218,37 @@ int main(int argc, char *argv[]){
       }
       
       /* This process has not completed.  Does it have any output? */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
       if(proc->eof or not FD_ISSET(proc->fd, &rfds)){ /* Spurious
 							 warning from
-							 -Wconversion */
+							 -Wconversion
+							 in GNU libc
+							 before
+							 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
 	/* This process had nothing to say at this time */
 	proc = proc->next;
 	continue;
       }
       /* Before reading, make the process' data buffer large enough */
       if(proc->buffer_length + BUFFER_SIZE > proc->buffer_size){
-	proc->buffer = realloc(proc->buffer, proc->buffer_size
-			       + (size_t) BUFFER_SIZE);
-	if(proc->buffer == NULL){
+	char *new_buffer = realloc(proc->buffer, proc->buffer_size
+				   + (size_t) BUFFER_SIZE);
+	if(new_buffer == NULL){
 	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
+	proc->buffer = new_buffer;
 	proc->buffer_size += BUFFER_SIZE;
       }
       /* Read from the process */
@@ -1253,8 +1307,10 @@ int main(int argc, char *argv[]){
     free(custom_argv);
   }
   
-  if(dir != NULL){
-    closedir(dir);
+  free(direntries);
+  
+  if(dir_fd != -1){
+    TEMP_FAILURE_RETRY(close(dir_fd));
   }
   
   /* Kill the processes */
