@@ -2,8 +2,8 @@
 /*
  * Password-prompt - Read a password from the terminal and print it
  * 
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -72,16 +72,6 @@ const char *argp_program_bug_address = "<mandos@recompile.se>";
 /* Needed for conflict resolution */
 const char plymouth_name[] = "plymouthd";
 
-__attribute__((format (gnu_printf, 2, 3), nonnull(1)))
-int fprintf_plus(FILE *stream, const char *format, ...){
-  va_list ap;
-  va_start (ap, format);
-  
-  TEMP_FAILURE_RETRY(fprintf(stream, "Mandos plugin %s: ",
-			     program_invocation_short_name));
-  return TEMP_FAILURE_RETRY(vfprintf(stream, format, ap));
-}
-
 /* Function to use when printing errors */
 __attribute__((format (gnu_printf, 3, 4)))
 void error_plus(int status, int errnum, const char *formatstring,
@@ -92,7 +82,7 @@ void error_plus(int status, int errnum, const char *formatstring,
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);
@@ -222,7 +212,7 @@ bool conflict_detection(void){
   struct dirent **direntries = NULL;
   int ret;
   ret = scandir("/proc", &direntries, is_plymouth, alphasort);
-  if (ret == -1){
+  if(ret == -1){
     error_plus(1, errno, "scandir");
   }
   free(direntries);
@@ -313,7 +303,7 @@ int main(int argc, char **argv){
     fprintf(stderr, "Starting %s\n", argv[0]);
   }
 
-  if (conflict_detection()){
+  if(conflict_detection()){
     if(debug){
       fprintf(stderr, "Stopping %s because of conflict\n", argv[0]);
     }
